import inspect
import pkgutil

from .module_router import ModuleRouter

"""
    PACKAGE EXTRACTOR
        It's purpose is to extract all the modules based on the paths of the parameter given
        it extracts valid packages within the root module path
        and register the valid modules to the flask blueprint
    :param application
        flask application object
            example: flask_app = Flask(import_name, instance_relative_config=True)
            application=flask_app
    :param paths
        paths are a list of directories within your application root folder
            example:
                paths = ['path/to/your/application/modules']
"""


class PackageExtractor:
    __packages = None
    __modules = []
    __routers = []

    def __init__(self, application, paths):
        self.paths = paths
        self.application = application
        self.__iter_paths()

    def __iter_paths(self):
        self.__extract_packages(packages=pkgutil.walk_packages(self.paths, prefix='', onerror=None))

    def __extract_packages(self, packages):
        if inspect.isgenerator(packages):
            try:
                loader, name, is_pkg = next(packages)
                self.__extract_modules(loader, name, is_pkg)
                self.__extract_packages(packages)
            except StopIteration:
                pass
        else:
            raise TypeError("package should be a generator type")

    def __serialize_module_paths(self):
        return list(set([path.split("/")[-1:][0] for path in self.paths]))

    """ extract modules from the package"""

    def __extract_modules(self, loader, name, is_pkg):

        """ if module found load module and save all attributes in the module found """
        mod = loader.find_module(name).load_module(name)

        """ find the attribute method on each module """
        if hasattr(mod, '__method__'):

            """ register to the blueprint if method attribute found """
            module_router = ModuleRouter(mod,
                                         ignore_names=self.__serialize_module_paths()
                                         ).register_route(app=self.application, name=name)

            self.__routers.extend(module_router.routers)
            self.__modules.append(mod)

        else:
            """ prompt not found notification """
            # print('{} has no module attribute method'.format(mod))
            pass

    @property
    def modules(self):
        return self.__modules

    @property
    def routers(self):
        return self.__routers



import inspect
from enum import Enum


class ErrorMessage(Enum):
    METHOD_NOT_A_LIST = 'Methods must be a type of list.'
    INVALID_ROUTE = 'Invalid route syntax'
    NAME_NOT_A_LIST = 'names must be a type of list'
    MEMBER_NOT_A_FUNCTION = 'Member is not a function.'


class PreDefinedHttpMethod(Enum):
    INDEX = 'index'
    CREATE = 'create'
    UPDATE = 'update'
    DESTROY = 'destroy'


class ModuleRouter:

    __routers = []
    __ignore_names = ["index", "routes"]

    def __init__(self, mod, **kwargs):  # module
        self._module = mod
        if "ignore_names" in kwargs:
            self.__ignore_names.extend(kwargs['ignore_names'])
        if self._is_valid_module():
            self.model_add_router()

    def model_add_router(self):
        if hasattr(self._module, "__routes__") and len(self._module.__routes__):
            route_type, route_data = self._routing_type(route=self._module.__routes__.pop(0))
            if route_type == 'cls':
                """ If it's a class it needs to extract the methods by function names
                    magic functions are excluded
                """
                route_name, slug, cls = route_data
                self.class_member_route(route=route_data, members=self.get_cls_fn_members(cls))

            elif route_type == 'fn':
                route_name, slug, fn, methods = route_data
                self.__routers.append(route_data)
                self._module.__method__.add_url_rule(
                    rule=slug,
                    endpoint=fn.__name__,
                    view_func=fn,
                    methods=methods)
            self.model_add_router()

    def _is_valid_module(self):
        return hasattr(self._module, "__routes__") or hasattr(self._module, "__method__")

    @staticmethod
    def _routing_type(route):
        __type = None
        if isinstance(route, tuple):
            if len(route) == 3 and inspect.isclass(route[2]):
                __type = 'cls'
            elif len(route) == 4 and inspect.isfunction(route[2]):
                if isinstance(route[3], (list, tuple, set)):
                    __type = 'fn'
                else:
                    raise TypeError(ErrorMessage.METHOD_NOT_A_LIST)
            else:
                raise TypeError(ErrorMessage.INVALID_ROUTE)
        return __type, route

    @staticmethod
    def get_http_methods(names):
        if isinstance(names, list):
            methods = []

            for name in names:
                if "__" not in name:
                    if name == PreDefinedHttpMethod.INDEX:
                        methods.append('GET')
                    elif name == PreDefinedHttpMethod.CREATE:
                        methods.append('POST')
                    elif name == PreDefinedHttpMethod.UPDATE:
                        methods.append('PUT')
                    elif name == PreDefinedHttpMethod.DESTROY:
                        methods.append('DELETE')
                    else:
                        methods.append('GET')

            return methods
        else:
            raise TypeError(ErrorMessage.NAME_NOT_A_LIST)

    @staticmethod
    def get_cls_fn_members(cls):
        return [member for member in inspect.getmembers(cls, predicate=inspect.isfunction)]

    def class_member_route(self, route, members):
        if isinstance(members, (list, set)):
            if len(members):
                (fn_name, fn_object) = members.pop(0)
                route_name, slug, cls = route
                if inspect.isfunction(fn_object):

                    _methods = self.get_http_methods([fn_name])

                    self._module.__method__.add_url_rule(
                        rule=slug,
                        endpoint=fn_name,
                        view_func=fn_object,
                        methods=_methods)
                    self.__routers.append((slug, fn_name, fn_object, _methods))
                else:
                    raise KeyError(ErrorMessage.MEMBER_NOT_A_FUNCTION)
                self.class_member_route(route, members)
        else:
            raise TypeError(ErrorMessage.MEMBER_NOT_A_FUNCTION)

    def register_route(self, app, name):
        app.register_blueprint(self._module.__method__, url_prefix=self.blueprint_name(name))
        return self

    def ignore_default_names(self, name):
        for ignore_name in list(set(self.__ignore_names)):
            if ignore_name in name:
                name = str(name).replace(ignore_name, "")
        return name

    def blueprint_name(self, name):
        """ set index automatically as home page """
        return self.blueprint_name_to_url(self.ignore_default_names(name))

    @staticmethod
    def blueprint_name_to_url(name):
        """ remove the last . in the string it it ends with a .
            for the url structure must follow the flask routing format
            it should be /model/method instead of /model/method/
        """
        if name[-1:] == ".":
            name = name[:-1]
        if name[-1:] is not "/":
            name = "/" + name
        name = str(name).replace(".", "/")
        return name

    @property
    def routers(self):
        return self.__routers



from distutils.core import setup
from flask_blueprint import __version__

setup(
    name='flask-blueprint',
    version=__version__,
    description='Flask blueprint generator',
    author='Jeffrey Marvin Forones',
    author_email='aiscenblue@gmail.com',
    url='https://github.com/aiscenblue/flask-blueprint',
    license='MIT',
    packages=['flask_blueprint'],
    keywords=['flask', 'blueprint', 'flask-blueprint', 'flask_blueprint'],  # arbitrary keywords
    install_requires=[
        # list of this package dependencies
    ],
    entry_points=None,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Framework :: Flask'
    ]
)



from .package_extractor import PackageExtractor

"""
    Description:: Initialize the blueprints inside in the root folder
    and sub folder

    Requirements:: all directories and sub directories must consist of __init__.py
    to be considered as a package. 

    files are ignored if its not end with .py or __.py

    NOTE :: directories must not consist of __ in their name

"""

__all__ = ['Core', '__version__']
__version__ = '1.2.8'


"""
    Core
        :param app
            your flask application module
                example: app = Flask('my_flask_app')
        
        :param root_paths
            paths for your module directories        

"""


class Core:
    __app = None

    def __init__(self, app, root_paths):

        """ save flask app module """
        self.__app = app
        self.root_path = root_paths

        """ register blueprint to the current path """
        PackageExtractor(application=app, paths=root_paths)
